@@ -1,38 +1,59 @@
 """ _base_events.py - event loop implementation using Qt """
 
 import asyncio
-import selectors
+import signal
 import sys
 import threading
 import traceback
 from asyncio import events
-from typing import Callable, List, Optional, Tuple
+from typing import Optional
+from ._selectable import *
 
 
-__all__ = ('AsyncSlotYield', 'AsyncSlotNotifier', 'AsyncSlotBaseEventLoop')
+__all__ = 'AsyncSlotBaseEventLoop',
 
 
-class AsyncSlotYield(Exception):
+class _AsyncSlotYield(Exception):
     """ Raised by an AsyncSlotSelectable to indicate that no IO is readily
     available and that _run_once should yield to the Qt event loop. """
     pass
 
 
-class AsyncSlotNotifier:
-    def notify(self) -> None:
-        raise NotImplementedError
+class _InterruptEvent:
+    __slots__ = '_flag',
 
-    def close(self) -> None:
-        raise NotImplementedError
+    def __init__(self):
+        self._flag = False
+
+    def set(self) -> None:
+        self._flag = True
+
+    def is_set(self) -> bool:
+        return self._flag
+
+    def clear(self) -> None:
+        self._flag = False
 
 
-AsyncSlotNotifierObject = None
+def _interrupt_handler(sig, frame):
+    assert sig == signal.SIGINT
+    # if frame is not None:
+    #     print(frame.f_locals)
+    if frame is not None and '_interrupt_event' in frame.f_locals:
+        _interrupt_event = frame.f_locals['_interrupt_event']
+        if isinstance(_interrupt_event, _InterruptEvent):
+            _interrupt_event.set()
+            return
+    return signal.default_int_handler(sig, frame)
 
 
-def create_notifier(callback: Callable[[], None]):
-    global AsyncSlotNotifierObject
-    if AsyncSlotNotifierObject is not None:
-        return AsyncSlotNotifierObject(callback)
+_AsyncSlotNotifierObject = None
+
+
+def _create_notifier(loop: "AsyncSlotBaseEventLoop"):
+    global _AsyncSlotNotifierObject
+    if _AsyncSlotNotifierObject is not None:
+        return _AsyncSlotNotifierObject(loop)
 
     from .bindings import QtCore
 
@@ -42,79 +63,72 @@ def create_notifier(callback: Callable[[], None]):
         else:
             _notified = QtCore.Signal()
 
-        def __init__(self, callback: Callable[[], None]):
+        def __init__(self, loop: "AsyncSlotBaseEventLoop"):
             super().__init__()
-            assert callback is not None, 'callback must not be None'
-            self._callback: Optional[Callable[[], None]] = callback
+            # The following creates a reference cycle.  Call close() to
+            # break the cycle.
+            self._loop = loop
             # Must make queued connection to avoid re-entrance.
             self._notified.connect(self._on_notified,
                                    QtCore.Qt.ConnectionType.QueuedConnection)
+            # Install a SIGINT handler if no custom handler is installed.
+            self._interrupt_handler_installed = False
+            if signal.getsignal(signal.SIGINT) is signal.default_int_handler:
+                try:
+                    signal.signal(signal.SIGINT, _interrupt_handler)
+                except (ValueError, OSError):
+                    pass
+                else:
+                    self._interrupt_handler_installed = True
 
-        def _on_notified(self):
-            if self._callback is not None:
-                self._callback()
-            else:
-                # TODO: print a warning that notification is received after
-                # TODO: the notifier is closed.
-                pass
+        def _on_notified(self, _interrupt_event=_InterruptEvent()):
+            # If Ctrl+C is pressed while the loop is in a 'non-blocking'
+            # select(), the select() will be woken up (due to set_wakeup_fd)
+            # and the _notified signal emitted.  KeyboardInterrupt will be
+            # raised at the first point where Python byte code is run, i.e.
+            # this method.  We wrap the body in try-except to handle this.
+            try:
+                if _interrupt_event.is_set():
+                    raise KeyboardInterrupt
+                if self._loop is not None:
+                    self._loop._asyncslot_loop_iteration()
+                else:
+                    # TODO: print a warning that notification is received after
+                    # TODO: the notifier is closed.
+                    pass
+            except KeyboardInterrupt as exc:
+                # We catch Ctrl+C only once.  If Ctrl+C is pressed again
+                # immediately, the program will crash.
+                if self._loop is not None:
+                    self._loop._asyncslot_loop_interrupt(exc)
+                else:
+                    pass  # ignore Ctrl+C for once
+            finally:
+                _interrupt_event.clear()
+
+        def no_result(self):
+            raise _AsyncSlotYield
 
         def notify(self):
             self._notified.emit()
 
+        def wakeup(self):
+            self._loop._write_to_self()
+
         def close(self):
-            if self._callback is not None:
+            if self._loop is not None:
                 self._notified.disconnect()
-                self._callback = None
+                self._loop = None
+            if self._interrupt_handler_installed:
+                try:
+                    if signal.getsignal(signal.SIGINT) is _interrupt_handler:
+                        signal.signal(signal.SIGINT, signal.default_int_handler)
+                except (ValueError, OSError):
+                    pass
+                finally:
+                    self._interrupt_handler_installed = False
 
-    AsyncSlotNotifierObject = _AsyncSlotNotifierObject
-    return AsyncSlotNotifierObject(callback)
-
-
-class AsyncSlotSelectable:
-    """Protocol for a 'selector' that supports non-blocking select and
-    notification.
-
-    A selector may be in one of the following states:
-      - IDLE   : the selector is not in BUSY or CLOSED state
-      - BUSY   : the last call to select() raised AsyncSlotYield, and
-                 a thread worker is waiting for IO or timeout
-      - CLOSED : close() has been called
-
-    State machine:
-      - [start] --- __init__ --> IDLE
-      - IDLE --- close() --> CLOSED
-        IDLE --- select
-                 - (IO ready, timeout == 0, or notifier is None) --> IDLE
-                 - (IO not ready, timeout != 0, and notifier not None) --> BUSY
-        IDLE --- set_notifier --> IDLE
-      - BUSY --- (IO ready or timeout reached) --> IDLE
-        BUSY --- set_notifier --> (wakes up selector) --> IDLE
-      - CLOSED --- [end]
-    """
-
-    def set_notifier(self, notifier: Optional[AsyncSlotNotifier]) -> None:
-        """Set the notifier.
-
-        If the selector is in BUSY state, wake it up and wait for it
-        to become IDLE before returning.  In this case, the previous
-        installed notifier (if any) is still signaled.
-        """
-        raise NotImplementedError
-
-    def select(self, timeout: Optional[float] = None) \
-            -> List[Tuple[selectors.SelectorKey, int]]:
-        """
-        If timeout is zero or some IO is readily available, return the
-        available IO immediately.
-
-        If timeout is not zero, IO is not ready and notifier is not None,
-        raise AsyncSlotYield and perform the real select() in a different
-        thread.  When that select() completes, signal the notifier object.
-
-        If timeout is not zero, IO is not ready and notifier is None,
-        perform normal (blocking) select.
-        """
-        raise NotImplementedError
+    return _AsyncSlotNotifierObject(loop)
 
 
 class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
@@ -186,11 +200,11 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
     Such clean-up code should finish as soon as possible, and should not
     access any Qt object as no Qt loop is running.
     """
-    def __init__(self, *args, standalone=True):
+    def __init__(self, *args, **kwargs):
         # If True, run_forever will launch the loop in STANDALONE mode.
         # If False, run_forever will launch the loop in EXCLUSIVE mode;
         # use start() to launch the loop in INTEGRATED mode.
-        self.__standalone = standalone
+        self.__standalone = True
 
         # If self is created in STANDALONE mode and is running,
         # __qt_event_loop is set to the QEventLoop that is being run.
@@ -201,9 +215,9 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         # is set to a notifier object and is attached to the selector to
         # notify us when IO is available or timeout occurs.  In any other
         # case, __notifier is set to None.
-        self.__notifier: Optional[AsyncSlotNotifier] = None
+        self.__notifier: Optional[_AsyncSlotNotifier] = None
 
-        # __processing is set to True in __process_asyncio_events to
+        # __processing is set to True in _asyncslot_loop_iteration to
         # indicate that a 'normal' asyncio event processing iteration
         # (i.e. _run_once) is running.  It is also set to True when the
         # loop is running in EXCLUSIVE mode.
@@ -229,12 +243,17 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         # Need to invoke base constructor after initializing member variables
         # for compatibility with Python 3.7's BaseProactorEventLoop (Windows),
         # which calls self.call_soon() indirectly from its constructor.
-        super().__init__(*args)  # noqa
+        super().__init__(*args, **kwargs)  # noqa
 
     # =========================================================================
     # Custom method for AsyncSlot
     # =========================================================================
 
+    def set_guest(self, guest: bool) -> None:
+        self._check_closed()
+        self._check_running()
+        self.__standalone = not guest
+
     def run_task(self, coro, *, name=None):
         try:
             self.__call_soon_eagerly = True
@@ -249,8 +268,8 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
 
     def start(self) -> None:
         if self.__standalone:
-            raise RuntimeError('AsyncSlotEventLoop.start() cannot be called '
-                               'for a loop created with standalone=True')
+            raise RuntimeError('AsyncSlotBaseEventLoop.start() can only be '
+                               'called for a loop operating in guest mode')
         self._asyncslot_loop_startup()
 
     def _asyncslot_loop_startup(self) -> None:
@@ -269,7 +288,7 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
 
         self.__old_agen_hooks = old_agen_hooks
 
-        self.__notifier = create_notifier(self.__process_asyncio_events)
+        self.__notifier = _create_notifier(self)
         self.__notifier.notify()  # schedule initial _run_once
 
         # Do not set notifier if a TestSelector is used (during testing).
@@ -299,7 +318,7 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         sys.set_asyncgen_hooks(*old_agen_hooks)
         # ---- END COPIED FROM BaseEventLoop.run_forever
 
-    def __process_asyncio_events(self):
+    def _asyncslot_loop_iteration(self):
         """ This slot is connected to the notified signal of self.__notifier,
         which is emitted whenever asyncio events are possibly available
         and need to be processed."""
@@ -312,26 +331,12 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         try:
             self.__processing = True
             self._run_once()  # defined in asyncio.BaseEventLoop
-        except AsyncSlotYield:
+        except _AsyncSlotYield:
             # Ignore _stopping flag until select() returns.  This follows
             # asyncio behavior.
             pass
         except BaseException as exc:
-            if self.__qt_event_loop is not None:
-                # In STANDALONE mode, propagate the exception to the caller
-                # of run_forever.
-                self.__run_once_error = exc
-                self.__qt_event_loop.exit(1)
-            else:
-                # In INTEGRATED mode, stop the loop immediately and raise
-                # the error into the Qt event loop.  For PyQt this will
-                # terminate the process; for PySide this will log an error
-                # and then ignored.
-                self._asyncslot_loop_cleanup()
-                raise
-                # TODO: implement consistent behavior under both bindings.
-                # TODO: maybe call some exception handler and let it decide
-                # TODO: what to do?
+            self._asyncslot_loop_interrupt(exc)
         else:
             # To be consistent with asyncio behavior, check the _stopping
             # flag only after running a full iteration of _run_once.
@@ -350,6 +355,37 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         finally:
             self.__processing = False
 
+    def _asyncslot_loop_interrupt(self, exc: BaseException):
+        """Terminate the loop abnormally with the given exception.
+
+        This method is called by:
+
+          (1) AsyncSlotBaseEventLoop._asyncslot_loop_iteration() if
+              _run_once raised a BaseException, typically SystemExit
+              or KeyboardInterrupt; and
+
+          (2) _AsyncSlotNotifier._on_notified() if KeyboardInterrupt
+              is raised during processing the notification.
+
+        """
+        if self.__qt_event_loop is not None:
+            # In STANDALONE mode, propagate the exception to the caller
+            # of run_forever.
+            self.__run_once_error = exc
+            self.__qt_event_loop.exit(1)
+        else:
+            # In INTEGRATED mode, stop the loop immediately and raise
+            # the error into the Qt event loop.  For PyQt this will
+            # terminate the process; for PySide this will log an error
+            # and then ignored.
+            # TODO: if KeyboardInterrupt is raised, the object may not
+            # TODO: be in a consistent state to perform cleanup.
+            self._asyncslot_loop_cleanup()
+            raise exc
+            # TODO: implement consistent behavior under both bindings.
+            # TODO: maybe call some exception handler and let it decide
+            # TODO: what to do?
+
     # =========================================================================
     # Compatibility with Python 3.7
     # =========================================================================
@@ -422,7 +458,7 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
 
         The precise semantics are as follows:
 
-        If the loop is created with standalone == True:
+        If the loop is operating in 'host' mode:
 
           - If stop() is called when the loop is STOPPED, the next time the
             loop is RUNNING it will run exactly one iteration and then stop.
@@ -434,7 +470,7 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
             loop is RUNNING, treat as if called via call_soon_threadsafe():
             wake up the selector, which will run one iteration and stop.
 
-        If the loop is created with standalone == False:
+        If the loop is operating in 'guest' mode:
 
           - If stop() is called when the loop is STOPPED, raise an error.
             (The pre-stop idiom is not supported.)
@@ -451,8 +487,8 @@ class AsyncSlotBaseEventLoop(asyncio.BaseEventLoop):
         In any case, if KeyboardInterrupt or SystemExit is raised when
         processing an iteration, that iteration is interrupted, the loop
         is stopped immediately, and the exception is propagated to the
-        caller of run_forever if running in STANDALONE or EXCLUSIVE mode
-        or thrown into the Qt loop if running in STANDALONE mode.
+        caller of run_forever if operating in 'host' mode or thrown into
+        the Qt loop if operating in 'guest' mode.
         """
         if self.__standalone:
             if self.is_running() and not self.__processing:
