@@ -3,42 +3,38 @@
 import asyncio.selector_events
 import concurrent.futures
 import selectors
+import signal
 import threading
 import unittest.mock
-import weakref
 from typing import List, Optional, Tuple
 from ._base_events import *
+from ._selectable import _AsyncSlotNotifier
 
 
 __all__ = 'AsyncSlotBaseSelectorEventLoop',
 
 
-class AsyncSlotSelector(selectors.BaseSelector):
+class _AsyncSlotSelector(selectors.BaseSelector):
 
-    def __init__(self, selector: selectors.BaseSelector,
-                 write_to_self: weakref.WeakMethod):
+    def __init__(self, selector: selectors.BaseSelector):
         super().__init__()
         self._selector = selector
-        self._write_to_self = write_to_self
         self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
         self._select_future: Optional[concurrent.futures.Future] = None
         self._idle = threading.Event()
         self._idle.set()
-        self._notifier: Optional[AsyncSlotNotifier] = None
+        self._notifier: Optional[_AsyncSlotNotifier] = None
         self._closed = False
 
-    def set_notifier(self, notifier: Optional[AsyncSlotNotifier]) -> None:
+    def set_notifier(self, notifier: Optional[_AsyncSlotNotifier]) -> None:
         self._unblock_if_blocked()
         self._notifier = notifier
 
     def _unblock_if_blocked(self):
         assert not self._closed, 'selector already closed'
         if not self._idle.is_set():
-            write_to_self = self._write_to_self()
-            assert write_to_self is not None, (
-                'AsyncSlotEventLoop is supposed to close AsyncSlotSelector '
-                'before being deleted')
-            write_to_self()
+            assert self._notifier is not None, 'notifier expected'
+            self._notifier.wakeup()
             self._idle.wait()
 
     def register(self, fileobj, events, data=None):
@@ -93,15 +89,15 @@ class AsyncSlotSelector(selectors.BaseSelector):
             self._idle.set()
             raise
         else:
-            raise AsyncSlotYield
+            return self._notifier.no_result()  # raises _AsyncSlotYield
 
     def _select(self, timeout):
-        # Make a copy of self._notifier because it may be altered by
-        # set_notifier immediately after self._idle is set.
-        notifier = self._notifier
         try:
             return self._selector.select(timeout)
         finally:
+            # Make a copy of self._notifier because it may be altered by
+            # set_notifier immediately after self._idle is set.
+            notifier = self._notifier
             self._idle.set()
             notifier.notify()
 
@@ -132,13 +128,41 @@ class AsyncSlotBaseSelectorEventLoop(
     AsyncSlotBaseEventLoop,
     asyncio.selector_events.BaseSelectorEventLoop
 ):
-    def __init__(self, selector=None, *, standalone=True):
+    def __init__(self, selector=None):
         if selector is None:
             selector = selectors.DefaultSelector()
         if isinstance(selector, unittest.mock.Mock):
             # Pass through mock object for testing
             asyncslot_selector = selector
         else:
-            asyncslot_selector = AsyncSlotSelector(
-                selector, weakref.WeakMethod(self._write_to_self))
-        super().__init__(asyncslot_selector, standalone=standalone)
+            asyncslot_selector = _AsyncSlotSelector(selector)
+        super().__init__(asyncslot_selector)
+
+        # Similar to asyncio.BaseProactorEventLoop, install wakeup fd
+        # so that select() in a separate thread can be interrupted by
+        # Ctrl+C.  Only the main thread of the main interpreter may
+        # install a wakeup fd, but other threads will never receive a
+        # KeyboardInterrupt, so it's ok if set_wakeup_fd fails.
+        self.__wakeup_fd_installed = False
+        self._asyncslot_install_wakeup_fd()
+
+    def _asyncslot_install_wakeup_fd(self):
+        try:
+            signal.set_wakeup_fd(self._csock.fileno())
+        except Exception:
+            self.__wakeup_fd_installed = False
+        else:
+            self.__wakeup_fd_installed = True
+
+    def close(self):
+        if self.is_running():
+            raise RuntimeError("Cannot close a running event loop")
+        # Uninstall the wakeup fd is one was installed.
+        if self.__wakeup_fd_installed:
+            try:
+                signal.set_wakeup_fd(-1)
+            except (ValueError, OSError):
+                pass
+            finally:
+                self.__wakeup_fd_installed = False
+        super().close()
